
/*****************************************************************
*  Lecteur de transpondeur animal à 134,2 KHz avec un ATMega 328 *
* et un afficheur 2 lignes 16 caractères grace à un lecteur de   *
* RFID EM4305-LF. Sur la réception d'une chaine hexadécimale de  *
* 13 bytes, on contrôle le Checksum (transmision intacte) le pays*
* et les données d'identification animale . Pour l'exemple :     *
* Données dumodule EM4305LF: 03 84 12 DB FA E7 D5                *                  
* Numéro d'identification du pays: 03 84                         *
* Numéro d'identification national: 12 DB FA E7 D5               *
* Protocole série: AA 0F 08 00 03 84 12 DB FA E7 D5 81 BB        *
* Code de départ: AA                                             *
* Code fixe: 0F 08 00                                            *
* Données de l'étiquette: 03 84 12 DB FA E7 D5                   *
* Code de contrôle(Checksum): 81                                 *      
* Code fin de trame: BB                                          *
* Calcul du code de contrôle: 0F(XOR)08(XOR)00(XOR)03(XOR)       *
* 84(XOR)12(XOR)DB(XOR) FA(XOR)E7(XOR) d5 = 81                   *
*   Read More: https://amazing-vacuum-tubes.com                  *
******************************************************************/

#include <SoftwareSerial.h>
#include <LiquidCrystal.h>

const int BUFFER_SIZE = 13;       // Reservation de 13 Bytes dans le Buffer
const int DATA_SIZE = 10;         // 10 Octets de données exploitées
const int DATA_FIXED_SIZE = 3;    // 3 bytes fixes constructeur (0F 08 00)
const int DATA_COUNTRY_SIZE = 2;  // 2 octets pour le pays  (example 03 84) = 900 dec
const int DATA_TAG_SIZE = 5;      // 5 octets N° (example 12 DB FA E7 D5)
const int CHECKSUM_SIZE = 1;      // 1 byte checksum (example 81)

SoftwareSerial ssrfid = SoftwareSerial(6, 7); // Rx, Tx (Tx non utilisé ici )

uint8_t buffer[BUFFER_SIZE]; // Pour stocker une trame de données entrante

int buffer_index = 0;

LiquidCrystal lcd(5, 4, 3, 2, 1, 0);  //Pin LCD: RS,E, D4-D7,RW=0V,

const int LedPin = 14;      // Cde rétro-eclairage


void setup() {
  lcd.begin(16,2); //Afficheur 2 lignes 16 caractères

  ssrfid.begin(9600); // Défini la vitesse de transmission  
  ssrfid.listen(); // en mode lecture

  pinMode(LedPin, OUTPUT); 

  digitalWrite(LedPin, HIGH); // Allume le rétroéclage de l'écran
  lcd.clear();                 // Efface l'écran
  lcd.print("   Lecteur RFID");  // Message de l'écran d'accueil
  lcd.setCursor(0, 1);
  lcd.print("Animal 134.2KHz" );
  delay (4000);  
  lcd.clear();                 
  lcd.print("amazing-vacuum-");
  lcd.setCursor(0, 1);
  lcd.print("   tubes.com");
  delay (4000);
  digitalWrite(LedPin, LOW); // Eteind le rétroéclage de l'écran
}

void loop() {

  if (ssrfid.available() > 0) {     // si les données sont présentes
    bool call_extract_tag = false;

    int ssvalue = ssrfid.read(); // lire des données

    if (ssvalue == -1) {         // aucune donnée n'a été lue
      return;
    }

    if (ssvalue == 170 && buffer_index == 0) {           // EM4305-LF a trouvé un transpondeur  => étiquette à lire
      digitalWrite(LedPin, HIGH);   // Allume le rétroéclage de l'écran
    } else if (ssvalue == 187 && buffer_index == BUFFER_SIZE - 1) {    // le message est complètement transmis
      call_extract_tag = true;      // extraire le mot à la fin 
    }

    if (buffer_index >= BUFFER_SIZE) {  // Controle débordement de tampon (si bégaiement !)
      lcd.clear();
      digitalWrite(LedPin, HIGH); // Allume le rétroéclage de l'écran
      lcd.print("Msg trop Long");
      buffer_index = 0;
      return;
    }

    buffer[buffer_index++] = ssvalue; // copier la valeur actuelle dans le tampon si tout est bon

    if (call_extract_tag == true) {
      if (buffer_index == BUFFER_SIZE) {
        extract_tag();
        buffer_index = 0;
      } else { // et si quelque chose ne va pas... remise à zéro du tampon et on recommence
        buffer_index = 0;
        return;
      }
    }
  }
}

void extract_tag() {         // analyse des données récupérées
  uint8_t msg_checksum = buffer[11]; 
  uint8_t msg_data[DATA_SIZE];
  uint8_t msg_data_country[DATA_COUNTRY_SIZE];
  uint8_t msg_data_tag[DATA_TAG_SIZE];
  uint8_t countrymessage[2 * DATA_COUNTRY_SIZE]; // stockage le code ASCII du pays après conversion des valeurs hexadécimales
  uint8_t tagmessage[2 * DATA_TAG_SIZE]; // même chose en code ASCII de la partie balise
  long checksum = 0;
  long country = 0;
  long long tag = 0;
  long s;                 
  int expo;               

  for (int i = 0; i < DATA_SIZE; i++) {
    msg_data[i] = buffer[i + 1];
    checksum ^= msg_data[i];              // a = a XOR b
  }

  
  for (int i = 0; i < DATA_COUNTRY_SIZE; i++) {
    msg_data_country[i] = buffer[i + 4];
  }

  // mettre le pays à l'envers dans le tampon de message du pays
  for (int i = 0; i < DATA_COUNTRY_SIZE ; i++) {
    String str = String(buffer[i + 4], HEX);
    if (str.length() == 1 && buffer[i + 6] - 48 < 10) {  // ajouter 0 en tête si < 10
      str = "0" + str;
    }
    str.toUpperCase();
    for (int k = 0; k < str.length(); k++) {
      byte x = str.charAt(k);
      countrymessage[2 * DATA_COUNTRY_SIZE - 1 - i * 2 - k] = x;
    }
  }

  country = hexInDec(countrymessage, 0, 2 * DATA_COUNTRY_SIZE);

  for (int i = 0; i < DATA_TAG_SIZE; i++) {
    msg_data_tag[i] = buffer[i + 6];
  }


  // Mettre le message à l'envers dans le tampon 
  for (int i = 0; i < DATA_TAG_SIZE ; i++) {
    String str = String(buffer[i + 6], HEX);
    if (str.length() == 1 && buffer[i + 6] - 48 < 10) {  // ajouter 0 en tête si < 10
      str = "0" + str;
    }
    str.toUpperCase();
    for (int k = 0; k < str.length(); k++) {
      byte x = str.charAt(k);
      tagmessage[2 * DATA_TAG_SIZE - 1 - i * 2 - k] = x;
    }
  }

  tag = hexInDec(tagmessage, 0, 2 * DATA_TAG_SIZE);


  s = long(tag % 1000000);


  lcd.clear();
  digitalWrite(LedPin, HIGH); // Allume le rétroéclage de l'écran
  lcd.print(country);
  if (checksum == msg_checksum) { // compare le checksum au checksum calculé
    lcd.print("Msg OK !"); // si checksum identique !
  } else {
    lcd.print(" (Msg non OK)"); // ou non
  }
  lcd.setCursor(0, 1);
  lcd.print(long(tag / 1000000));
  s = long(tag % 1000000);
  for (long i = 100000; s < i && i > 1; i = i / 10) { // met un 0 devant
    lcd.print("0");
  }
  lcd.print(long(tag % 1000000));

}

long long hexInDec(char message[], int beg , int len) {
  long long mult = 1;
  long long nbr = 0;
  byte nextInt;
  for (int i = beg; i < beg + len; i++) {
    nextInt = message[i];
    if (nextInt >= 48 && nextInt <= 57) nextInt = map(nextInt, 48, 57, 0, 9);
    if (nextInt >= 65 && nextInt <= 70) nextInt = map(nextInt, 65, 70, 10, 15);
    if (nextInt >= 97 && nextInt <= 102) nextInt = map(nextInt, 97, 102, 10, 15);
    nextInt = constrain(nextInt, 0, 15);
    nbr = nbr + (mult * nextInt);
    mult = mult * 16;
  }
  return nbr;
}
